from collections import namedtuple

import numpy as np


class DelayData:
    _Delay = namedtuple("Delay", ("values", "errors"))

    def __init__(self, readback_address, setpoint_address, encoder_address):
        self._readback_address = readback_address
        self._setpoint_address = setpoint_address
        self._encoder_address = encoder_address
        self._is_ready_for_new_data = None
        self._is_ready_for_new_scan_point = None
        self._scan_range = None
        self._scan_step_size = None
        self._samples_per_step = None
        self._scan_step = None
        self._direction = None
        self._delays = []
        self._delay = None

    @property
    def readback_address(self):
        return self._readback_address

    @property
    def setpoint_address(self):
        return self._setpoint_address

    @property
    def encoder_address(self):
        return self._encoder_address

    @property
    def is_ready_for_new_data(self):
        return self._is_ready_for_new_data

    @property
    def is_ready_for_new_scan_point(self):
        return self._is_ready_for_new_scan_point

    @property
    def delay(self):
        return self._delay

    def initialize(self, scan_range, samples_per_step):
        self._scan_range = scan_range
        self._scan_step_size = np.abs(np.diff(self._scan_range)[0])
        self._samples_per_step = samples_per_step
        self._is_ready_for_new_scan_point = True
        self._scan_step = 0
        self._delays.clear()
        self._delay = None

    def reset(self):
        self._is_ready_for_new_scan_point = False
        self._is_ready_for_new_data = False

    def setup_scan_range(self, current_delay):
        if np.any(current_delay > self._scan_range) & np.any(current_delay < self._scan_range):
            self._scan_step = np.abs(self._scan_range - current_delay).argmin()
            self._direction = np.sign(np.diff(self._scan_range)[0])
        elif np.abs(self._scan_range - current_delay).argmin() != 0:
            self._scan_range = np.flip(self._scan_range)
            self._direction = -1
        else:
            self._direction = 1

    def is_ready_for_new_setpoint(self):
        if self._is_ready_for_new_scan_point:
            self._is_ready_for_new_scan_point = False
            self._is_ready_for_new_data = False
            return True
        return False

    def check_delay(self, readback_timing, setpoint_timing):
        if readback_timing == setpoint_timing:
            self._is_ready_for_new_data = True

    def get_next_setpoint(self):
        if self._scan_step == len(self._scan_range):
            self._scan_step = 0
            self._scan_range = np.flip(self._scan_range)
            self._direction *= -1
        next_setpoint = self._scan_range[self._scan_step]
        self._scan_step += 1
        return next_setpoint

    def get_scanning_progress(self):
        return self._direction * self._scan_step / len(self._scan_range) * 100

    def update(self, delay_data):
        self._delays.append(delay_data)
        if len(self._delays) % self._samples_per_step == 0:
            self._is_ready_for_new_scan_point = True

    def get_indices_and_bins(self):
        data = self._delays
        bins = int(np.floor(round((np.nanmax(data) - np.nanmin(data)) / self._scan_step_size, 3)) + 1)
        delta = np.nanmin(data) + self._scan_step_size * bins - np.nanmax(data)
        boundaries = np.linspace(np.nanmin(data) - delta / 2, np.nanmax(data) + delta / 2, bins + 1)
        return np.searchsorted(boundaries, data), bins

    def process_with(self, process_function):
        delay = process_function(self._delays)
        self._delay = self._Delay(delay[:, 0], delay[:, 1])
