import time
from collections import namedtuple

import numpy as np


class EnergyData:
    _Energy = namedtuple("Energy", ("values", "errors"))

    def __init__(self, readback_address, setpoint_address, setpoint_change_address):
        self._readback_address = readback_address
        self._setpoint_address = setpoint_address
        self._setpoint_change_address = setpoint_change_address
        self._readback_energy = None
        self._is_energy_to_set = False
        self._is_ready_for_new_data = None
        self._is_ready_for_new_scan_point = None
        self._energy = None
        self._scan_range = None
        self._samples_per_step = None
        self._tolerance = None
        self._timeout = None
        self._scan_step = None
        self._last_time_energy_to_set = None
        self._energies = []

    @property
    def readback_address(self):
        return self._readback_address

    @property
    def setpoint_address(self):
        return self._setpoint_address

    @property
    def setpoint_change_address(self):
        return self._setpoint_change_address

    @property
    def readback_energy(self):
        return self._readback_energy

    @property
    def is_energy_to_set(self):
        return self._is_energy_to_set

    @property
    def is_ready_for_new_data(self):
        return self._is_ready_for_new_data

    @property
    def is_ready_for_new_scan_point(self):
        return self._is_ready_for_new_scan_point

    @property
    def scan_step(self):
        return self._scan_step

    @property
    def energy(self):
        return self._energy

    def initialize(self, scan_range, samples_per_step, tolerance, timeout):
        self._scan_range = scan_range
        self._samples_per_step = samples_per_step
        self._tolerance = tolerance
        self._timeout = timeout
        self._is_energy_to_set = True
        self._is_ready_for_new_scan_point = True
        self._scan_step = 0
        self._energies.clear()
        self._energy = None

    def reset(self):
        self._is_energy_to_set = False
        self._is_ready_for_new_scan_point = False
        self._is_ready_for_new_data = False

    def is_ready_for_new_setpoint(self):
        if self._is_ready_for_new_scan_point:
            self._last_time_energy_to_set = time.time()
            self._is_ready_for_new_scan_point = False
            self._is_ready_for_new_data = False
            return True
        return False

    def get_next_setpoint(self):
        next_setpoint = self._scan_range[self._scan_step]
        self._scan_step += 1
        return next_setpoint

    def check_energy(self, readback_energy, setpoint_energy):
        delta_energy = abs(readback_energy - setpoint_energy)
        delta_time = time.time() - self._last_time_energy_to_set
        if delta_energy <= self._tolerance or delta_time >= self._timeout:
            self._is_ready_for_new_data = True

    def update(self, energy_data):
        self._readback_energy = energy_data
        if self._is_ready_for_new_data:
            self._energies.append(energy_data)
            if len(self._energies) % self._samples_per_step == 0:
                self._is_ready_for_new_scan_point = True
                energy = np.array(np.split(np.array(self._energies), self._scan_step))
                energy_values_and_errors = np.stack((np.nanmean(energy, axis=1), np.nanstd(energy, axis=1)), axis=-1)
                self._energy = self._Energy(energy_values_and_errors[:, 0], energy_values_and_errors[:, 1])

    def get_scanning_progress(self):
        return len(self._energies) / (self._samples_per_step * len(self._scan_range)) * 100

    def is_not_empty(self):
        return self._energy is not None
