from collections import namedtuple

import numpy as np


class ImageData:
    _Projections = namedtuple("Projections", ("horizontal", "vertical"))
    _Values = namedtuple("Values", ("x", "y"))
    _Units = namedtuple("Units", ("horizontal", "vertical"))

    class ImageSizeChangedWhileTakingBackground(Exception):
        def __init__(self):
            super().__init__("Image size changed while taking background")

    class ImageSizeChangedWhileRunning(Exception):
        def __init__(self):
            super().__init__("Image size changed while running")

    class ImageSizeChangedWhileAveraging(Exception):
        def __init__(self):
            super().__init__("Image size changed while averaging")

    class _ImageSizeChanged(Exception):  # marker exception
        pass

    @classmethod
    def get_relevant_image_and_offset(cls, image):
        raise NotImplementedError

    def __init__(self, address):
        self._address = address
        self._image = None
        self._projections_values = None
        self._projections_units = self._Units("px", "px")
        self._is_averaging = None
        self._average_samples = 0
        self._average_image = None
        self._is_taking_background = None
        self._background_samples = None
        self._background_sample = None
        self._background_image = None
        self._is_correcting_background = None
        self._corrected_image = None

    @property
    def address(self):
        return self._address

    @property
    def image(self):
        return self._image

    @property
    def projections_values(self):
        return self._projections_values

    @property
    def projections_units(self):
        return self._projections_units

    @property
    def average_samples(self):
        return self._average_samples

    @property
    def is_taking_background(self):
        return self._is_taking_background

    def set_averaging(self, state):
        self._is_averaging = state
        self.reset_averaging()

    def reset_averaging(self):
        self._average_samples = 0

    def set_background_taking(self, background_samples):
        self._is_taking_background = True
        self._background_samples = background_samples
        self._background_sample = 0

    def reset_background_taking(self):
        self._is_taking_background = False

    def set_background_correction(self, state):
        self._is_correcting_background = state

    def update(self, image_data):
        self._image = image_data
        self._update_image()
        self.update_projections()

    def _update_image(self):
        if self._is_taking_background:
            self._apply_background_taking()
        else:
            if self._is_correcting_background:
                self._apply_background_correction()
            if self._is_averaging:
                self._apply_image_averaging()

    def _apply_background_taking(self):
        try:
            self._background_sample, self._background_image = self._averaging(
                self._background_sample, self._background_image, self._image)
        except self._ImageSizeChanged:
            raise self.ImageSizeChangedWhileTakingBackground() from None
        finally:
            self._image = self._background_image

    def _apply_background_correction(self):
        if self._image.shape == self._background_image.shape:
            self._corrected_image = self._image - self._background_image
            self._image = self._corrected_image
        else:
            if self._corrected_image is None:
                self._image = self._background_image
            else:
                self._image = self._corrected_image
            raise self.ImageSizeChangedWhileRunning()

    def _apply_image_averaging(self):
        try:
            self._average_samples, self._average_image = self._averaging(
                self._average_samples, self._average_image, self._image)
        except self._ImageSizeChanged:
            raise self.ImageSizeChangedWhileAveraging() from None
        finally:
            self._image = self._average_image

    @classmethod
    def _averaging(cls, samples, average, image):
        samples += 1
        if samples == 1:
            average = image
        elif image.shape == average.shape:
            inverse_samples = 1 / samples  # use this to avoid memory issue due to implicit casting
            average = (1 - inverse_samples) * average + inverse_samples * image
        else:
            raise cls._ImageSizeChanged()
        return samples, average

    def update_projections(self):
        image, offset = self.get_relevant_image_and_offset(self._image)
        self._projections_values = self._Projections(self._calculate_projection(image, offset, axis=0),
                                                     self._calculate_projection(image, offset, axis=1))

    def _calculate_projection(self, image, offset, axis):
        return self._Values(np.arange(0, len(np.nansum(image, axis=axis)), dtype=np.float) + offset[axis],
                            np.nansum(image, axis=axis) / 1e3)   # divided by 1e3 to have a proper value range

    def is_not_empty(self):
        return self._image is not None

    def get_background_taking_progress(self):
        return self._background_sample / self._background_samples * 100
