import numpy as np
from doocspie.util.color import Colors
from pyqtgraph import GraphicsLayoutWidget, ErrorBarItem, PlotDataItem, LegendItem, QtCore


class ScannerPlotter:
    _DESY_ORANGE = np.array(Colors.DESY.ORANGE) * 255
    _DESY_CYAN = np.array(Colors.DESY.CYAN) * 255
    _DESY_GREY = np.array(Colors.DESY.GRAY_1) * 255

    def __init__(self):
        self._widget = GraphicsLayoutWidget()

        self._scan_plot = self._widget.addPlot()
        self._scan_error = ErrorBarItem(x=[], y=[])
        self._scan_line = PlotDataItem()
        self._scan_fit = PlotDataItem()

        self._scan_plot.addItem(self._scan_error)
        self._scan_plot.addItem(self._scan_line)
        self._scan_plot.addItem(self._scan_fit)

        self._legend = LegendItem(offset=(70, 15))
        self._current_legend = None

        self._initialize_plots()

    def _initialize_plots(self):
        self._scan_plot.setMenuEnabled(False)
        self._scan_plot.showAxis("top", show=True)
        self._scan_plot.showAxis("right", show=True)
        for pos in "top", "right", "left", "bottom":
            self._scan_plot.getAxis(pos).setStyle(showValues=False, tickLength=0)
        self._scan_error.setData(pen={"color": self._DESY_GREY, "width": 2})
        self._scan_line.setData(symbol="o", pen=None,
                                symbolBrush=tuple(self._DESY_ORANGE), symbolPen=tuple(self._DESY_ORANGE))
        self._scan_fit.setData(pen={"color": self._DESY_CYAN, "width": 2})

    @property
    def widget(self):
        return self._widget

    def set_scan(self, energy, metrics_values, metrics_errors, label, metrics_unit):
        self._scan_error.setData(x=energy.values, y=np.array(metrics_values),
                                 width=2*energy.errors, height=2*np.array(metrics_errors))
        self._scan_line.setData(x=energy.values, y=metrics_values)

        self._scan_plot.getAxis("left").setStyle(showValues=True)
        self._scan_plot.setLabel("left", f"{label} ({metrics_unit})")
        self._scan_plot.getAxis("bottom").setStyle(showValues=True)
        self._scan_plot.setLabel("bottom", "Readback Energy (eV)")

    def set_fit(self, fit):
        self._scan_fit.setData(x=fit.energy, y=fit.metrics)
        if self._current_legend != fit.legend:
            self._legend.removeItem(self._current_legend)
            self._current_legend = fit.legend
            self._legend.setParentItem(self._scan_plot)
            self._legend.addItem(self._scan_fit, fit.legend)
            self._legend.setGeometry(QtCore.QRectF(0, 0, self._legend.width() + 10, self._legend.height()))

    def clear_fit(self):
        self._scan_fit.clear()
        self._legend.removeItem(self._current_legend)
        self._current_legend = None
        if self._legend.scene():
            self._legend.scene().removeItem(self._legend)

    def clear(self):
        self._scan_error.setData(x=[], y=[], width=np.array([]), height=np.array([]))
        self._scan_line.clear()
        self._scan_fit.clear()
        self._scan_plot.setLabel("left", "")
        self._scan_plot.getAxis("left").setStyle(showValues=False, tickLength=0)
        self._scan_plot.setLabel("bottom", "")
        self._scan_plot.getAxis("bottom").setStyle(showValues=False, tickLength=0)
