import numpy as np

#from . import doocsmy as doocspie


import doocspie


class IOService:
    class IOException(Exception):
        def __init__(self, error):
            super().__init__("I/O error: " + error)

    class NoCameraImageType(Exception):
        def __init__(self):
            super().__init__("No camera image type")

    @classmethod
    def get(cls, address, meta_event=None):
        try:
            return doocspie.get(address, meta_event=meta_event)
        except Exception as err:
            raise cls.IOException(str(err))

    @classmethod
    def set(cls, address, data):
        try:
            doocspie.set(address, data)
        except Exception as err:
            raise cls.IOException(str(err))

    def __init__(self, image_data, energy_data, position_data):
        self._image_data = image_data
        self._energy_data = energy_data
        self._position_data = position_data
        self._image_readout = None
        self._event = None
        self._is_new_data_available = None

    @property
    def image_data(self):
        return self._image_data

    @property
    def energy_data(self):
        return self._energy_data

    @property
    def position_data(self):
        return self._position_data

    @property
    def is_new_data_available(self):
        return self._is_new_data_available

    def update(self):
        self._update_energy_setpoint()
        self._update_position_setpoint()
        self._update_data()

    def _update_energy_setpoint(self):
        if self._energy_data.is_energy_to_set:
            self._set_energy()
            self._check_energy()

    def _set_energy(self):
        if self._energy_data.is_ready_for_new_setpoint():
            self.set(self._energy_data.setpoint_address, self._energy_data.get_next_setpoint())
            self.set(self._energy_data.setpoint_change_address, 1)

    def _check_energy(self):
        if not self._energy_data.is_ready_for_new_data:
            self._energy_data.check_energy(self.get(self._energy_data.readback_address).data,
                                           self.get(self._energy_data.setpoint_address).data)

    def _update_position_setpoint(self):
        if self._position_data.is_position_to_set:
            self._set_position()
            self._check_position()

    def _set_position(self):
        if self._position_data.is_ready_for_new_setpoint():
            self._position_data.set_reference(self.get(self._position_data.reference_address).data)
            next_scan_point = self._position_data.get_next_setpoint()
            if next_scan_point == self.get(self._position_data.setpoint_address).data:
                self._position_data.set_ready_to_check()
            else:
                self.set(self._position_data.setpoint_address, next_scan_point)

        if not self._position_data.is_ready_to_check_position:
            if self._position_data.reference != self.get(self._position_data.reference_address).data:
                self._position_data.set_ready_to_check()
                self.set(self._position_data.setpoint_change_address, 1)

    def _check_position(self):
        if self._position_data.is_ready_to_check_position and not self._position_data.is_ready_for_new_data:
            self._position_data.check_position(self.get(self._position_data.readback_address).data,
                                               self.get(self._position_data.setpoint_address).data)

    def _update_data(self):
        self._image_readout = self.get(self._image_data.address)
        if self._image_readout.type != "IMAGE":
            raise self.NoCameraImageType()
        self._is_new_data_available = False
        if self._image_readout.event != self._event:
            self._image_data.update(self._image_readout.data.astype(np.float32))
            self._energy_data.update(self.get(self._energy_data.readback_address).data)
            self._position_data.update(self.get(self._position_data.readback_address).data)
            self._event = self._image_readout.event
            self._is_new_data_available = True
