from PyQt5.QtWidgets import (QMainWindow, QVBoxLayout, QPushButton, QSizePolicy, QCheckBox, QRadioButton, QWidget,
                             QHBoxLayout, QSplitter, QGroupBox)

from ..plotter.logger_plotter import LoggerPlotter


class LoggerToolUi(QMainWindow):
    START_LABEL = "Start"

    _CENTER_LABEL = "Center"
    _WIDTH_LABEL = "RMS Width"
    _PEAK_LABEL = "Peak"

    def __init__(self):
        super().__init__()
        self._name = "Logger"
        self._main_layout = QVBoxLayout()
        self._logger_plotter = LoggerPlotter(self._CENTER_LABEL, self._WIDTH_LABEL, self._PEAK_LABEL)
        self._start_button = self._get_labeled_push_button(self.START_LABEL)
        self._stop_button = self._get_labeled_push_button("Stop")
        self._reset_button = self._get_labeled_push_button("Reset")
        self._enable_movable_line_checkbox = QCheckBox("Enable Line Marker", checked=True)
        self._show_fit_curve_checkbox = QCheckBox("Show Fit Curve", checked=True)
        self._horizontal_center_radio_button = QRadioButton(self._CENTER_LABEL, checked=True)
        self._horizontal_peak_radio_button = QRadioButton(self._PEAK_LABEL)
        self._horizontal_width_radio_button = QRadioButton(self._WIDTH_LABEL)
        self._vertical_center_radio_button = QRadioButton(self._CENTER_LABEL, checked=True)
        self._vertical_peak_radio_button = QRadioButton(self._PEAK_LABEL)
        self._vertical_width_radio_button = QRadioButton(self._WIDTH_LABEL)
        self._create_layout()
        self._create_connections()

    @staticmethod
    def _get_labeled_push_button(label):
        push_button = QPushButton(label)
        push_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        return push_button

    def _create_layout(self):
        self._create_main_layout()
        self._create_plot_layout()
        self._create_controls_layout()

    def _create_main_layout(self):
        main_widget = QWidget()
        main_widget.setLayout(self._main_layout)
        self.setCentralWidget(main_widget)

    def _create_plot_layout(self):
        plot_layout = QHBoxLayout()
        splitter = QSplitter()
        splitter.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        splitter.addWidget(self._logger_plotter.horizontal_widget)
        splitter.addWidget(self._logger_plotter.vertical_widget)
        plot_layout.addWidget(splitter)
        self._main_layout.addLayout(plot_layout)

    def _create_controls_layout(self):
        controls_layout = QHBoxLayout()
        controls_layout.addWidget(self._create_control_group_box())
        controls_layout.addWidget(self._create_horizontal_metrics_group_box())
        controls_layout.addWidget(self._create_vertical_metrics_group_box())
        self._main_layout.addLayout(controls_layout)

    def _create_control_group_box(self):
        group_box = QGroupBox("Control")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._start_button)
        widgets_layout.addWidget(self._stop_button)
        widgets_layout.addWidget(self._reset_button)
        widgets_layout.addWidget(self._enable_movable_line_checkbox)
        widgets_layout.addWidget(self._show_fit_curve_checkbox)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_horizontal_metrics_group_box(self):
        group_box = QGroupBox("Horizontal Parameters")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._horizontal_center_radio_button)
        widgets_layout.addWidget(self._horizontal_peak_radio_button)
        widgets_layout.addWidget(self._horizontal_width_radio_button)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_vertical_metrics_group_box(self):
        group_box = QGroupBox("Vertical Parameters")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._vertical_center_radio_button)
        widgets_layout.addWidget(self._vertical_peak_radio_button)
        widgets_layout.addWidget(self._vertical_width_radio_button)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_connections(self):
        self._reset_button.clicked.connect(self._logger_plotter.reset)
        self._enable_movable_line_checkbox.stateChanged.connect(self._logger_plotter.set_movable_line_visible)
        self._show_fit_curve_checkbox.stateChanged.connect(self._logger_plotter.set_fit_curve_visible)
        self._horizontal_center_radio_button.clicked.connect(
            lambda: self._logger_plotter.set_horizontal_movable_line(self._horizontal_center_radio_button.text()))
        self._horizontal_width_radio_button.clicked.connect(
            lambda: self._logger_plotter.set_horizontal_movable_line(self._horizontal_width_radio_button.text()))
        self._horizontal_peak_radio_button.clicked.connect(
            lambda: self._logger_plotter.set_horizontal_movable_line(self._horizontal_peak_radio_button.text()))
        self._vertical_center_radio_button.clicked.connect(
            lambda: self._logger_plotter.set_vertical_movable_line(self._vertical_center_radio_button.text()))
        self._vertical_width_radio_button.clicked.connect(
            lambda: self._logger_plotter.set_vertical_movable_line(self._vertical_width_radio_button.text()))
        self._vertical_peak_radio_button.clicked.connect(
            lambda: self._logger_plotter.set_vertical_movable_line(self._vertical_peak_radio_button.text()))

    @property
    def name(self):
        return self._name

    @property
    def start_button(self):
        return self._start_button

    @property
    def stop_button(self):
        return self._stop_button

    @property
    def reset_button(self):
        return self._reset_button

    @property
    def is_fit_visible(self):
        return self._logger_plotter.is_fit_visible

    def add_to_connections_for_changes(self, action):
        self._horizontal_center_radio_button.clicked.connect(action)
        self._horizontal_width_radio_button.clicked.connect(action)
        self._horizontal_peak_radio_button.clicked.connect(action)
        self._vertical_center_radio_button.clicked.connect(action)
        self._vertical_width_radio_button.clicked.connect(action)
        self._vertical_peak_radio_button.clicked.connect(action)

    def plot_projections(self, values, units):
        self._logger_plotter.set_projections(values, units)

    def plot_fits(self, fits):
        self._logger_plotter.set_fits(fits)

    def plot_metrics(self, values, units):
        self._plot_horizontal_metrics(values.horizontal, units.horizontal)
        self._plot_vertical_metrics(values.vertical, units.vertical)

    def _plot_horizontal_metrics(self, values, units):
        if self._horizontal_center_radio_button.isChecked():
            self._logger_plotter.set_horizontal_metrics(values.center, units.center,
                                                        self._horizontal_center_radio_button.text())
        elif self._horizontal_width_radio_button.isChecked():
            self._logger_plotter.set_horizontal_metrics(values.rms, units.rms,
                                                        self._horizontal_width_radio_button.text())
        elif self._horizontal_peak_radio_button.isChecked():
            self._logger_plotter.set_horizontal_metrics(values.peak, units.peak,
                                                        self._horizontal_peak_radio_button.text())

    def _plot_vertical_metrics(self, values, units):
        if self._vertical_center_radio_button.isChecked():
            self._logger_plotter.set_vertical_metrics(values.center, units.center,
                                                      self._vertical_center_radio_button.text())
        elif self._vertical_width_radio_button.isChecked():
            self._logger_plotter.set_vertical_metrics(values.rms, units.rms,
                                                      self._vertical_width_radio_button.text())
        elif self._vertical_peak_radio_button.isChecked():
            self._logger_plotter.set_vertical_metrics(values.peak, units.peak,
                                                      self._vertical_peak_radio_button.text())

    def get_selected_horizontal_metrics(self, metrics):
        if self._horizontal_center_radio_button.isChecked():
            return metrics.metrics_values.horizontal.center
        if self._horizontal_peak_radio_button.isChecked():
            return metrics.metrics_values.horizontal.peak
        if self._horizontal_width_radio_button.isChecked():
            return metrics.metrics_values.horizontal.rms
        return None

    def get_selected_vertical_metrics(self, metrics):
        if self._vertical_center_radio_button.isChecked():
            return metrics.metrics_values.vertical.center
        if self._vertical_peak_radio_button.isChecked():
            return metrics.metrics_values.vertical.peak
        if self._vertical_width_radio_button.isChecked():
            return metrics.metrics_values.vertical.rms
        return None

    def get_selected_horizontal_label(self):
        if self._horizontal_center_radio_button.isChecked():
            return "horizontal center"
        if self._horizontal_peak_radio_button.isChecked():
            return "horizontal peak"
        if self._horizontal_width_radio_button.isChecked():
            return "horizontal r.m.s. width"
        return None

    def get_selected_vertical_label(self):
        if self._vertical_center_radio_button.isChecked():
            return "vertical center"
        if self._vertical_peak_radio_button.isChecked():
            return "vertical peak"
        if self._vertical_width_radio_button.isChecked():
            return "vertical r.m.s. width"
        return None
