from doocspie.pyqt import messages

from .pulse_settings_tool_ui import PulseSettingsToolUi


class PulseSettingsTool:

    def __init__(self, timing_service):
        self._pulse_settings_tool_ui = PulseSettingsToolUi()
        self._pulse_settings_tool_ui.closeEvent = self._close_event
        self._timing_service = timing_service
        self._is_initialized = False
        self._create_connections()

    def _close_event(self, event):
        try:
            self._timing_service.reflectivity_data.set_parameters(signal_range=self._get_signal_range(),
                                                                  background_range=self._get_background_range())
        except ValueError:
            event.ignore()

    def _create_connections(self):
        self._pulse_settings_tool_ui.threshold_slider.valueChanged.connect(self._update_threshold)

    @property
    def ui(self):
        return self._pulse_settings_tool_ui

    @property
    def is_initialized(self):
        return self._is_initialized

    def reset(self):
        self._is_initialized = False

    def initialize(self):
        if not self._is_initialized:
            self._timing_service.update_reflectivity_data()
            self._setup_reflectivity_data()

    def show_dialog(self):
        self.update()
        self._pulse_settings_tool_ui.show_window()

    def _update_threshold(self):
        self._setup_reflectivity_data()
        self._plot_and_set_pulses()

    def update(self):
        self._timing_service.update_reflectivity_data()
        self._setup_reflectivity_data()
        self._plot_and_set_pulses()

    def _setup_reflectivity_data(self):
        if not self._is_initialized:
            self._pulse_settings_tool_ui.threshold_slider.setValue(5)
            self._timing_service.reflectivity_data.set_parameters(signal_range=self._get_signal_range(),
                                                                  background_range=self._get_background_range())
        self._timing_service.reflectivity_data.find_threshold(self._pulse_settings_tool_ui.threshold_slider.value())
        self._pulse_settings_tool_ui.threshold_slider.setMinimum(
            self._timing_service.reflectivity_data.minimum_threshold_factor)
        self._pulse_settings_tool_ui.threshold_slider.setMaximum(
            self._timing_service.reflectivity_data.maximum_threshold_factor)
        self._timing_service.reflectivity_data.find_pulses()
        self._is_initialized = True

    def _get_signal_range(self):
        try:
            begin = int(self._pulse_settings_tool_ui.signal_begin_line_edit.text())
        except ValueError as err:
            messages.show_error("Empty/Faulty Signal 'Begin' field")
            raise err
        try:
            end = int(self._pulse_settings_tool_ui.signal_end_line_edit.text())
        except ValueError as err:
            messages.show_error("Empty/Faulty Signal 'End' field")
            raise err

        if begin > end:
            messages.show_error("Signal 'Begin' cannot be larger then 'End'")
            raise ValueError

        return begin, end

    def _get_background_range(self):
        try:
            begin = int(self._pulse_settings_tool_ui.background_begin_line_edit.text())
        except ValueError as err:
            messages.show_error("Empty/Faulty Background 'Begin' field")
            raise err
        try:
            end = int(self._pulse_settings_tool_ui.background_end_line_edit.text())
        except ValueError as err:
            messages.show_error("Empty/Faulty Background 'End' field")
            raise err

        if begin > end:
            messages.show_error("Background 'Begin' cannot be larger then 'End'")
            raise ValueError

        return begin, end

    def _plot_and_set_pulses(self):
        self._plot()
        self._set_pulses()

    def _plot(self):
        self._pulse_settings_tool_ui.plot_adc(self._timing_service.reflectivity_data.adc)
        self._pulse_settings_tool_ui.plot_threshold(self._timing_service.reflectivity_data.threshold)
        self._pulse_settings_tool_ui.plot_positions(self._timing_service.reflectivity_data.positions)

    def _set_pulses(self):
        self._pulse_settings_tool_ui.set_laser_pulses(self._timing_service.reflectivity_data.laser_pulses)
        self._pulse_settings_tool_ui.set_fel_pulses(self._timing_service.reflectivity_data.fel_pulses)
        self._pulse_settings_tool_ui.set_utilized_pulses(self._timing_service.reflectivity_data.utilized_pulses)
