from collections import namedtuple

import numpy as np
from doocspie.pyqt import messages

from .scanner_tool_ui import ScannerToolUi


class ScannerTool:
    _Fit = namedtuple("Fit", ["energy", "metrics", "legend"])

    def __init__(self, energy_data, metrics_data, fit_service):
        self._scanner_tool_ui = ScannerToolUi()
        self._energy_data = energy_data
        self._metrics_data = metrics_data
        self._fit_service = fit_service
        self._fit_model = None
        self._fit = None
        self._legend = None
        self._is_started = False
        self._create_connections()

    def _create_connections(self):
        self._scanner_tool_ui.fit_group.buttonClicked.connect(self._update_fit_model)
        self._scanner_tool_ui.start_button.clicked.connect(self._start_button_clicked)
        self._scanner_tool_ui.stop_button.clicked.connect(self.stop_button_clicked)
        self._scanner_tool_ui.add_to_connections_for_changes(self.update_scan)

    @property
    def ui(self):
        return self._scanner_tool_ui

    @property
    def is_started(self):
        return self._is_started

    def _update_fit_model(self):
        if self._scanner_tool_ui.fit_group.checkedButton():
            self._fit_model = self._fit_service.get_fit_model(self._scanner_tool_ui.fit_group.checkedButton().text())
        else:
            self._fit_model = None
        self.update_scan()

    def _start_button_clicked(self):
        if self._is_input_validation_successful():
            self._is_started = True
            self._scanner_tool_ui.start_button.setText("running ...")
            self._enable_control_and_input_fields(False)
            self._energy_data.initialize(scan_range=self._get_scan_range(),
                                         samples_per_step=int(self._scanner_tool_ui.samples_per_step_line_edit.text()),
                                         tolerance=int(self._scanner_tool_ui.tolerance_line_edit.text()) / 1000,
                                         timeout=float(self._scanner_tool_ui.timeout_line_edit.text()))
            self._metrics_data.initialize_scanner_metrics()
            self._scanner_tool_ui.clear()

    def _is_input_validation_successful(self):
        try:
            begin = float(self._scanner_tool_ui.begin_line_edit.text())
            if begin == 0:
                raise ValueError
        except ValueError:
            messages.show_error("Empty/Faulty/Zero 'Begin' field")
            return False
        try:
            end = float(self._scanner_tool_ui.end_line_edit.text())
            if end == 0:
                raise ValueError
        except ValueError:
            messages.show_error("Empty/Faulty/Zero 'End' field")
            return False
        try:
            int(self._scanner_tool_ui.number_of_steps_line_edit.text())
        except ValueError:
            messages.show_error("Empty 'Number of Steps' field")
            return False
        try:
            int(self._scanner_tool_ui.samples_per_step_line_edit.text())
        except ValueError:
            messages.show_error("Empty 'Samples / Step' field")
            return False
        try:
            int(self._scanner_tool_ui.tolerance_line_edit.text())
        except ValueError:
            messages.show_error("Empty 'Tolerance' field")
            return False
        try:
            float(self._scanner_tool_ui.timeout_line_edit.text())
        except ValueError:
            messages.show_error("Empty/Faulty 'Timeout' field")
            return False
        return True

    def _enable_control_and_input_fields(self, state):
        self._scanner_tool_ui.start_button.setEnabled(state)
        self._scanner_tool_ui.begin_line_edit.setEnabled(state)
        self._scanner_tool_ui.end_line_edit.setEnabled(state)
        self._scanner_tool_ui.number_of_steps_line_edit.setEnabled(state)
        self._scanner_tool_ui.samples_per_step_line_edit.setEnabled(state)
        self._scanner_tool_ui.tolerance_line_edit.setEnabled(state)
        self._scanner_tool_ui.timeout_line_edit.setEnabled(state)

    def _get_scan_range(self):
        energy_range_begin = float(self._scanner_tool_ui.begin_line_edit.text())
        energy_range_end = float(self._scanner_tool_ui.end_line_edit.text())
        steps = int(self._scanner_tool_ui.number_of_steps_line_edit.text())
        return np.linspace(energy_range_begin, energy_range_end, steps)

    def stop_button_clicked(self):
        if self._is_started:
            self._scanner_tool_ui.set_progress(0)
            self._scanner_tool_ui.start_button.setText(ScannerToolUi.START)
            self._enable_control_and_input_fields(True)
            self._energy_data.reset()
            self._is_started = False

    def update_scanning_progress(self):
        if self._is_started:
            self._scanner_tool_ui.set_progress(self._energy_data.get_scanning_progress())
            if self._energy_data.get_scanning_progress() == 100:
                self._scanner_tool_ui.stop_button.click()

    def update(self):
        if self._energy_data.is_ready_for_new_scan_point:
            self._plot()

    def _plot(self):
        self._scanner_tool_ui.plot_scan(self._energy_data.energy,
                                        self._metrics_data.scanner_metrics_values,
                                        self._metrics_data.scanner_metrics_errors,
                                        self._metrics_data.metrics_units)
        self._plot_fit()

    def _plot_fit(self):
        if self._fit_model:
            metrics_values = self._scanner_tool_ui.get_selected_metrics(self._metrics_data.scanner_metrics_values)
            metrics_errors = self._scanner_tool_ui.get_selected_metrics(self._metrics_data.scanner_metrics_errors)

            self._fit, self._legend = self._fit_service.get_fit_for(
                self._fit_model, self._energy_data.energy.values, np.array(metrics_values), np.array(metrics_errors))

            if self._fit is not None and self._legend is not None:
                self._scanner_tool_ui.plot_fit(self._Fit(self._fit[0], self._fit[1], self._legend))
            else:
                self._scanner_tool_ui.clear_fit()
        else:
            self._fit = None
            self._legend = None
            self._scanner_tool_ui.clear_fit()

    def update_scan(self):
        if self._energy_data.is_not_empty():
            self._plot()

    def get_export_content(self):
        energy_ev = "energy (eV)"
        energy_error_ev = "energy error (eV)"
        parameters_px = "parameters (px)"
        parameters_error_px = "parameters error (px)"
        fit_model = "fit model"
        fit_energy_ev = "fit energy (eV)"
        fit_parameters_px = "fit parameters (px)"
        fit_legend = "fit legend"

        content = {energy_ev: [], energy_error_ev: [], parameters_px: [], parameters_error_px: [],
                   fit_model: [], fit_energy_ev: [], fit_parameters_px: [], fit_legend: [],
                   "parameters label": [self._scanner_tool_ui.get_selected_label()],
                   "energy begin (eV)": [self._scanner_tool_ui.begin_line_edit.text()],
                   "energy end (eV)": [self._scanner_tool_ui.end_line_edit.text()],
                   "number of steps": [self._scanner_tool_ui.number_of_steps_line_edit.text()],
                   "samples per step": [self._scanner_tool_ui.samples_per_step_line_edit.text()],
                   "tolerance (meV)": [self._scanner_tool_ui.tolerance_line_edit.text()],
                   "timeout (s)": [self._scanner_tool_ui.timeout_line_edit.text()]}

        if self._energy_data.is_not_empty():
            content[energy_ev] = self._energy_data.energy.values
            content[energy_error_ev] = self._energy_data.energy.errors
            content[parameters_px] = self._scanner_tool_ui.get_selected_metrics(
                self._metrics_data.scanner_metrics_values)
            content[parameters_error_px] = self._scanner_tool_ui.get_selected_metrics(
                self._metrics_data.scanner_metrics_errors)
            if self._fit is not None:
                content[fit_model] = [self._fit_model.__name__]
                content[fit_energy_ev] = self._fit[0]
                content[fit_parameters_px] = self._fit[1]
            if self._legend is not None:
                content[fit_legend] = [self._legend.replace("&#177;", "\u00b1")]  # html -> unicode for "+-"

        return content
