from PyQt5.QtCore import QRegExp, Qt
from PyQt5.QtGui import QRegExpValidator
from PyQt5.QtWidgets import (QMainWindow, QVBoxLayout, QPushButton, QProgressBar, QLineEdit, QSizePolicy, QWidget,
                             QHBoxLayout, QGroupBox, QLabel, QRadioButton, QCheckBox, QButtonGroup)

from ..plotter.scanner_plotter import ScannerPlotter


class ScannerToolUi(QMainWindow):
    START = "Start"

    _CENTER_LABEL = "Center"
    _WIDTH_LABEL = "R.M.S. Width"
    _PEAK_LABEL = "Peak"

    def __init__(self):
        super().__init__()
        self._name = "Scanner"
        self._main_layout = QVBoxLayout()
        self._scanner_plotter = ScannerPlotter()
        self._start_button = self._get_push_button_with(self.START)
        self._stop_button = self._get_push_button_with("Stop")
        self._progress_bar = self._get_progress_bar()
        self._begin_line_edit = self._get_line_edit()
        self._end_line_edit = self._get_line_edit()
        self._number_of_steps_line_edit = self._get_line_edit_with("10", "[1-9][0-9]*")
        self._samples_per_step_line_edit = self._get_line_edit_with("10", "[1-9][0-9]*")
        self._tolerance_line_edit = self._get_line_edit_with("10", "[0-9]*")
        self._timeout_line_edit = self._get_line_edit_with("5.0", "[0-9]*[.][0-9]*")
        self._parameters_group = QButtonGroup(self)
        self._horizontal_center_radio_button = QRadioButton(self._CENTER_LABEL, checked=True)
        self._horizontal_peak_radio_button = QRadioButton(self._PEAK_LABEL)
        self._horizontal_width_radio_button = QRadioButton(self._WIDTH_LABEL)
        self._vertical_center_radio_button = QRadioButton(self._CENTER_LABEL)
        self._vertical_peak_radio_button = QRadioButton(self._PEAK_LABEL)
        self._vertical_width_radio_button = QRadioButton(self._WIDTH_LABEL)
        self._fit_group = QButtonGroup(self, exclusive=False)
        self._constant_check_box = QCheckBox("Constant")
        self._linear_check_box = QCheckBox("Linear")
        self._quadratic_check_box = QCheckBox("Quadratic")
        self._check_boxes = self._constant_check_box, self._linear_check_box, self._quadratic_check_box
        self._fit_model_state = 0
        self._create_layout()
        self._create_connections()

    @staticmethod
    def _get_push_button_with(label):
        push_button = QPushButton(label)
        push_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        return push_button

    @staticmethod
    def _get_progress_bar():
        progress_bar = QProgressBar()
        progress_bar.setValue(0)
        return progress_bar

    @staticmethod
    def _get_line_edit():
        line_edit = QLineEdit()
        line_edit.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        line_edit.setAlignment(Qt.AlignRight)
        line_edit.setValidator(QRegExpValidator(QRegExp("[0-9]*[.][0-9]*")))
        line_edit.setMaxLength(7)
        line_edit.setMaximumWidth(70)
        return line_edit

    @staticmethod
    def _get_line_edit_with(default_value, regex):
        integer_line_edit = QLineEdit(default_value)
        integer_line_edit.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        integer_line_edit.setAlignment(Qt.AlignRight)
        integer_line_edit.setMaxLength(4)
        integer_line_edit.setMaximumWidth(40)
        integer_line_edit.setValidator(QRegExpValidator(QRegExp(regex)))
        return integer_line_edit

    def _create_layout(self):
        self._create_main_layout()
        self._create_plot_layout()
        self._create_controls_layout()
        self._create_progress_layout()

    def _create_main_layout(self):
        main_widget = QWidget()
        main_widget.setLayout(self._main_layout)
        self.setCentralWidget(main_widget)

    def _create_plot_layout(self):
        plot_layout = QHBoxLayout()
        plot_layout.addWidget(self._scanner_plotter.widget)
        plot_layout.addLayout(self._create_analysis_layout())
        self._main_layout.addLayout(plot_layout)

    def _create_analysis_layout(self):
        analysis_layout = QVBoxLayout()
        analysis_layout.addWidget(self._create_parameters_group_box())
        analysis_layout.addWidget(self._create_fit_group_box())
        return analysis_layout

    def _create_parameters_group_box(self):
        group_box = QGroupBox("Parameters")
        widgets_layout = QVBoxLayout()
        widgets_layout.addWidget(self._create_horizontal_parameters_group_box())
        widgets_layout.addWidget(self._create_vertical_parameters_group_box())
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_horizontal_parameters_group_box(self):
        group_box = QGroupBox("Horizontal")
        widgets_layout = QVBoxLayout()
        widgets_layout.addWidget(self._horizontal_center_radio_button)
        widgets_layout.addWidget(self._horizontal_peak_radio_button)
        widgets_layout.addWidget(self._horizontal_width_radio_button)
        self._parameters_group.addButton(self._horizontal_center_radio_button)
        self._parameters_group.addButton(self._horizontal_peak_radio_button)
        self._parameters_group.addButton(self._horizontal_width_radio_button)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_vertical_parameters_group_box(self):
        group_box = QGroupBox("Vertical")
        widgets_layout = QVBoxLayout()
        widgets_layout.addWidget(self._vertical_center_radio_button)
        widgets_layout.addWidget(self._vertical_peak_radio_button)
        widgets_layout.addWidget(self._vertical_width_radio_button)
        self._parameters_group.addButton(self._vertical_center_radio_button)
        self._parameters_group.addButton(self._vertical_peak_radio_button)
        self._parameters_group.addButton(self._vertical_width_radio_button)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_fit_group_box(self):
        group_box = QGroupBox("Fit")
        widgets_layout = QVBoxLayout()
        widgets_layout.addWidget(self._create_fit_models_group_box())
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_fit_models_group_box(self):
        group_box = QGroupBox("Models")
        widgets_layout = QVBoxLayout()
        widgets_layout.addWidget(self._constant_check_box)
        widgets_layout.addWidget(self._linear_check_box)
        widgets_layout.addWidget(self._quadratic_check_box)
        self._fit_group.addButton(self._constant_check_box)
        self._fit_group.addButton(self._linear_check_box)
        self._fit_group.addButton(self._quadratic_check_box)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_controls_layout(self):
        controls_layout = QHBoxLayout()
        controls_layout.addWidget(self._create_control_group_box())
        controls_layout.addWidget(self._create_energy_range_group_box())
        controls_layout.addWidget(self._create_scan_setup_group_box())
        self._main_layout.addLayout(controls_layout)

    def _create_control_group_box(self):
        group_box = QGroupBox("Control")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(self._start_button)
        widgets_layout.addWidget(self._stop_button)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_energy_range_group_box(self):
        group_box = QGroupBox("Energy Range")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(QLabel("Begin (eV):"))
        widgets_layout.addWidget(self._begin_line_edit)
        widgets_layout.addWidget(QLabel("End (eV):"))
        widgets_layout.addWidget(self._end_line_edit)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_scan_setup_group_box(self):
        group_box = QGroupBox("Scan Setup")
        widgets_layout = QHBoxLayout()
        widgets_layout.addWidget(QLabel("Number of Steps:"))
        widgets_layout.addWidget(self._number_of_steps_line_edit)
        widgets_layout.addWidget(QLabel("Samples / Step:"))
        widgets_layout.addWidget(self._samples_per_step_line_edit)
        widgets_layout.addWidget(QLabel("Tolerance (meV):"))
        widgets_layout.addWidget(self._tolerance_line_edit)
        widgets_layout.addWidget(QLabel("Timeout (s):"))
        widgets_layout.addWidget(self._timeout_line_edit)
        group_box.setLayout(widgets_layout)
        return group_box

    def _create_progress_layout(self):
        progress_layout = QHBoxLayout()
        progress_layout.addWidget(QLabel("Progress:"))
        progress_layout.addWidget(self._progress_bar)
        self._main_layout.addLayout(progress_layout)

    def _create_connections(self):
        self._fit_group.buttonClicked.connect(self._update_fit_menu)

    def _update_fit_menu(self):
        fit_model_state = 0
        for index, check_box in enumerate(self._check_boxes):
            if check_box.isChecked():
                fit_model_state += 2 ** index

        if fit_model_state:
            self._fit_model_state ^= fit_model_state  # make sure only one fit model option is selected
        else:
            self._fit_model_state = fit_model_state  # or none at all

        for index, check_box in enumerate(self._check_boxes):  # set the fit model ui accordingly
            if 2 ** index & self._fit_model_state:
                check_box.setChecked(True)
            else:
                check_box.setChecked(False)

    @property
    def name(self):
        return self._name

    @property
    def start_button(self):
        return self._start_button

    @property
    def stop_button(self):
        return self._stop_button

    @property
    def progress_bar(self):
        return self._progress_bar

    @property
    def begin_line_edit(self):
        return self._begin_line_edit

    @property
    def end_line_edit(self):
        return self._end_line_edit

    @property
    def number_of_steps_line_edit(self):
        return self._number_of_steps_line_edit

    @property
    def samples_per_step_line_edit(self):
        return self._samples_per_step_line_edit

    @property
    def tolerance_line_edit(self):
        return self._tolerance_line_edit

    @property
    def timeout_line_edit(self):
        return self._timeout_line_edit

    @property
    def fit_group(self):
        return self._fit_group

    def add_to_connections_for_changes(self, action):
        self._horizontal_center_radio_button.clicked.connect(action)
        self._horizontal_peak_radio_button.clicked.connect(action)
        self._horizontal_width_radio_button.clicked.connect(action)
        self._vertical_center_radio_button.clicked.connect(action)
        self._vertical_peak_radio_button.clicked.connect(action)
        self._vertical_width_radio_button.clicked.connect(action)

    def set_progress(self, value):
        self._progress_bar.setValue(value)

    def plot_scan(self, energy, metrics_values, metrics_errors, metrics_units):
        if self._horizontal_center_radio_button.isChecked():
            self._scanner_plotter.set_scan(energy, metrics_values.horizontal.center, metrics_errors.horizontal.center,
                                           self._horizontal_center_radio_button.text(), metrics_units.horizontal.center)
        elif self._horizontal_peak_radio_button.isChecked():
            self._scanner_plotter.set_scan(energy, metrics_values.horizontal.peak, metrics_errors.horizontal.peak,
                                           self._horizontal_peak_radio_button.text(), metrics_units.horizontal.peak)
        elif self._horizontal_width_radio_button.isChecked():
            self._scanner_plotter.set_scan(energy, metrics_values.horizontal.rms, metrics_errors.horizontal.rms,
                                           self._horizontal_width_radio_button.text(), metrics_units.horizontal.rms)
        elif self._vertical_center_radio_button.isChecked():
            self._scanner_plotter.set_scan(energy, metrics_values.vertical.center, metrics_errors.vertical.center,
                                           self._vertical_center_radio_button.text(), metrics_units.vertical.center)
        elif self._vertical_peak_radio_button.isChecked():
            self._scanner_plotter.set_scan(energy, metrics_values.vertical.peak, metrics_errors.vertical.peak,
                                           self._vertical_peak_radio_button.text(), metrics_units.vertical.peak)
        elif self._vertical_width_radio_button.isChecked():
            self._scanner_plotter.set_scan(energy, metrics_values.vertical.rms, metrics_errors.vertical.rms,
                                           self._vertical_width_radio_button.text(), metrics_units.vertical.rms)

    def clear(self):
        self._scanner_plotter.clear()
        self.clear_fit()

    def get_selected_metrics(self, metrics):
        if self._horizontal_center_radio_button.isChecked():
            return metrics.horizontal.center
        if self._horizontal_peak_radio_button.isChecked():
            return metrics.horizontal.peak
        if self._horizontal_width_radio_button.isChecked():
            return metrics.horizontal.rms
        if self._vertical_center_radio_button.isChecked():
            return metrics.vertical.center
        if self._vertical_peak_radio_button.isChecked():
            return metrics.vertical.peak
        if self._vertical_width_radio_button.isChecked():
            return metrics.vertical.rms
        return None

    def plot_fit(self, fit):
        self._scanner_plotter.set_fit(fit)

    def clear_fit(self):
        self._scanner_plotter.clear_fit()

    def get_selected_label(self):
        if self._horizontal_center_radio_button.isChecked():
            return "horizontal center"
        if self._horizontal_peak_radio_button.isChecked():
            return "horizontal peak"
        if self._horizontal_width_radio_button.isChecked():
            return "horizontal r.m.s. width"
        if self._vertical_center_radio_button.isChecked():
            return "vertical center"
        if self._vertical_peak_radio_button.isChecked():
            return "vertical peak"
        if self._vertical_width_radio_button.isChecked():
            return "vertical r.m.s. width"
        return None
