from doocspie.pyqt import messages

from .viewer_tool_ui import ViewerToolUi


class ViewerTool:

    def __init__(self, image_data, energy_data, position_data, fit_service, roi_data):
        self._viewer_tool_ui = ViewerToolUi(roi_data)
        self._image_data = image_data
        self._image_data.get_relevant_image_and_offset = self._viewer_tool_ui.get_relevant_image_and_offset
        self._energy_data = energy_data
        self._position_data = position_data
        self._fit_service = fit_service
        self._is_background_correction_enabled = False
        self._is_averaging_selected = self._viewer_tool_ui.averaging_check_box.isChecked()
        self._is_started = False
        self._create_connections()

    def _create_connections(self):
        self._viewer_tool_ui.start_button.clicked.connect(self.start_button_clicked)
        self._viewer_tool_ui.take_button.clicked.connect(self._take_button_clicked)
        self._viewer_tool_ui.stop_button.clicked.connect(self.stop_button_clicked)
        self._viewer_tool_ui.averaging_check_box.stateChanged.connect(self._image_data.set_averaging)
        self._viewer_tool_ui.apply_background_correction_check_box.stateChanged.connect(
            self._image_data.set_background_correction)
        self._viewer_tool_ui.add_to_connections_for_changes(self._offline_update_projections)

    @property
    def ui(self):
        return self._viewer_tool_ui

    @property
    def is_started(self):
        return self._is_started

    def start_button_clicked(self):
        self._is_started = True
        self._viewer_tool_ui.start_button.setText("running ...")
        self._viewer_tool_ui.start_button.setEnabled(False)
        self._viewer_tool_ui.take_button.setEnabled(False)

    def stop_button_clicked(self):
        if self._is_started:
            self._image_data.reset_background_taking()
            self._viewer_tool_ui.samples_line_edit.setEnabled(True)
            self._viewer_tool_ui.take_button.setEnabled(True)
            self._viewer_tool_ui.show_take_button()
            self._viewer_tool_ui.start_button.setText(self._viewer_tool_ui.START_LABEL)
            self._viewer_tool_ui.start_button.setEnabled(True)
            self._is_started = False

    def _take_button_clicked(self):
        if self._is_input_validation_successful():
            self._is_started = True
            self._viewer_tool_ui.start_button.setEnabled(False)
            self._viewer_tool_ui.show_progress_bar()
            self._viewer_tool_ui.samples_line_edit.setEnabled(False)
            self._image_data.set_background_taking(int(self._viewer_tool_ui.samples_line_edit.text()))
            self.set_apply_background_correction_check_box_to(False)

    def _is_input_validation_successful(self):
        try:
            int(self._viewer_tool_ui.samples_line_edit.text())
        except ValueError:
            self.set_apply_background_correction_check_box_to(False)
            messages.show_error("Empty 'Samples' field")
            return False
        return True

    def update(self):
        self._viewer_tool_ui.plot_image(self._image_data.image)
        self._plot_and_show()

    def _plot_and_show(self):
        self._viewer_tool_ui.plot_projections(self._image_data.projections_values, self._image_data.projections_units)
        self._viewer_tool_ui.plot_fits(self._fit_service.fit_data.fits)
        self._viewer_tool_ui.show_parameters(self._fit_service.fit_data.parameters,
                                             self._image_data.average_samples,
                                             self._energy_data.readback_energy,
                                             self._position_data.readback_position)

    def _offline_update_projections(self):
        if not self._is_started and self._image_data.is_not_empty():
            self._image_data.update_projections()
            self._fit_service.update()
            self._plot_and_show()
            self._viewer_tool_ui.update_axis()

    def update_background_taking_progress(self):
        if self._image_data.is_taking_background:
            self._image_data.reset_averaging()
            self._viewer_tool_ui.set_progress(self._image_data.get_background_taking_progress())
            if self._image_data.get_background_taking_progress() == 100:
                self.set_apply_background_correction_check_box_to(True)
                self._viewer_tool_ui.stop_button.click()

    def offline_update(self):
        if self._image_data.is_not_empty():
            self.update()

    def set_apply_background_correction_check_box_to(self, state):
        self._is_background_correction_enabled = state
        self._viewer_tool_ui.apply_background_correction_check_box.setChecked(state)
        self._viewer_tool_ui.apply_background_correction_check_box.setEnabled(state)

    def recover_apply_background_correction_enabled_state(self):
        self._viewer_tool_ui.apply_background_correction_check_box.setEnabled(self._is_background_correction_enabled)

    def switch_off_averaging(self):
        self._viewer_tool_ui.averaging_check_box.setEnabled(False)
        is_averaging_selected = self._viewer_tool_ui.averaging_check_box.isChecked()
        self._viewer_tool_ui.averaging_check_box.setChecked(False)
        self._is_averaging_selected = is_averaging_selected  # using the state before setting the checkbox to false

    def recover_averaging(self):
        if not self._viewer_tool_ui.averaging_check_box.isEnabled():
            self._viewer_tool_ui.averaging_check_box.setEnabled(True)
            self._viewer_tool_ui.averaging_check_box.setChecked(self._is_averaging_selected)
